/*
 * shape.c --
 *
 *	This module implements raw access to the X Shaped-Window extension.
 *
 * Copyright (c) 1997 by Donal K. Fellows
 *
 * The author hereby grants permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 * $Id: shape.c,v 1.1 1997/09/14 22:08:32 donal Exp donal $
 */ 

#include <tk.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include "panic.h"

static int
applyOperationToToplevelParent = 1;

static Window
getWindow(apptkwin, interp, pathName, isToplevel)
     Tk_Window apptkwin;
     Tcl_Interp *interp;
     Tcl_Obj *pathName;
     int *isToplevel;
{
  Tk_Window tkwin =
    Tk_NameToWindow(interp, Tcl_GetStringFromObj(pathName, NULL), apptkwin);

  if (Tk_Display(tkwin) != Tk_Display(apptkwin)) {
    Tcl_AppendResult(interp, "can only apply shape operations to windows"
		     " on the same display as the main window of the"
		     " application", NULL);
    return None;
  }
  if (Tk_WindowId(tkwin) == None) {
    Tk_MakeWindowExist(tkwin);
    if (Tk_WindowId(tkwin) == None) {
      panic("bizarre failure to create window");
    }
  }
  *isToplevel = Tk_IsTopLevel(tkwin);
  return Tk_WindowId(tkwin);
}

static Window
getXParent(dpy, w)
     Display *dpy;
     Window w;
{
  Window root, parent, *kids;
  unsigned int nkids;

  if (XQueryTree(dpy, w, &root, &parent, &kids, &nkids)) {
    if (kids) {
      XFree(kids);
    }
    if (parent!=root) {
      return parent;
    }
  }
  return None;
}

static int
foo(clientData, interp, objc, objv)
     ClientData clientData;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  Display *dpy = Tk_Display((Tk_Window)clientData);

  static char *subcommands[] = {
    "offset", "rectangles", "version", NULL
  };
  enum { SHAPE_OFFSET, SHAPE_RECTS, SHAPE_VERSION };
  int subcmdidx;

  static char *kinds[] = { "bounding", "clip", "both", NULL };

  static char *operations[] = {
    "set", "union", "intersect", "subtract", "invert",
    ":=", "+=", "*=", "-=", "=!", NULL
  };
  static int opmap[] = {
    ShapeSet, ShapeUnion, ShapeIntersect, ShapeSubtract, ShapeInvert,
    ShapeSet, ShapeUnion, ShapeIntersect, ShapeSubtract, ShapeInvert
  };

  if (objc<2) {
    Tcl_WrongNumArgs(interp, 1, objv, "subcommand ?window arg ...?");
    return TCL_ERROR;
  } else if (Tcl_GetIndexFromObj(interp, objv[1], subcommands, "subcommand",
				 0, &subcmdidx) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (subcmdidx) {
  case SHAPE_OFFSET:
    /* shape offset .foo clip 12 34 */

    if (objc!=6) {
      Tcl_WrongNumArgs(interp, 1, objv, "offset pathName kind xOff yOff");
      return TCL_ERROR;
    } else {
      int isTop;
      Window w = getWindow(clientData, interp, objv[2], &isTop);
      int kind, xOff, yOff;

      if (w == None || Tcl_GetIndexFromObj(interp, objv[3], kinds,
					   "kind", 0, &kind) != TCL_OK ||
	  Tcl_GetIntFromObj(interp, objv[4], &xOff) != TCL_OK ||
	  Tcl_GetIntFromObj(interp, objv[5], &yOff) != TCL_OK) {
	return TCL_ERROR;
      }
      if (kind==0||kind==2) {
	XShapeOffsetShape(dpy, w, ShapeBounding,	/* What's going on */
			  xOff, yOff);			/* How are we moving */
	if (isTop && applyOperationToToplevelParent) {
	  Window parent = getXParent(dpy, w);

	  if (parent!=None) {
	    XShapeOffsetShape(dpy, parent, ShapeBounding, xOff, yOff);
	  }
	}
      }
      if (kind==1||kind==2) {
	XShapeOffsetShape(dpy, w, ShapeClip, xOff, yOff);
	if (isTop && applyOperationToToplevelParent) {
	  Window parent = getXParent(dpy, w);

	  if (parent!=None) {
	    XShapeOffsetShape(dpy, parent, ShapeClip, xOff, yOff);
	  }
	}
      }
      return TCL_OK;
    }

  case SHAPE_RECTS:
    /* shape rect .foo bound ?union rect ...? */
    if (objc<4 || objc==5) {
      Tcl_WrongNumArgs(interp, 1, objv,
		       "rectangles pathName kind ?operation rectangle ...?");
      return TCL_ERROR;
    } else {
      int isTop;
      Window w = getWindow(clientData, interp, objv[2], &isTop);
      int kind, op;

      if (w == None || Tcl_GetIndexFromObj(interp, objv[3], kinds,
					   "kind", 0, &kind) != TCL_OK) {
	return TCL_ERROR;
      }
      if (objc==4) {
	XRectangle *rects = NULL;
	int count,order,i;
	Tcl_Obj *rect[4], **retvals;

	switch (kind) {
	case 0:
	  rects = XShapeGetRectangles(dpy, w, ShapeBounding,	/* What's up */
				      &count,			/* Wieviel? */
				      &order);			/* ignored */
	  break;
	case 1:
	  rects = XShapeGetRectangles(dpy, w, ShapeClip, &count, &order);
	  break;
	case 2:
	  Tcl_AppendResult(interp, "can only query the rectangle list for "
			   "one kind of shaped region at a time", NULL);
	  return TCL_ERROR;
	}

	if (count) {
	  retvals = (Tcl_Obj **)Tcl_Alloc(sizeof(Tcl_Obj *)*count);
	  for (i=0 ; i<count ; i++) {
	    rect[0] = Tcl_NewIntObj(rects[i].x);
	    rect[1] = Tcl_NewIntObj(rects[i].y);
	    rect[2] = Tcl_NewIntObj(rects[i].x + rects[i].width  - 1);
	    rect[3] = Tcl_NewIntObj(rects[i].y + rects[i].height - 1);
	    retvals[i] = Tcl_NewListObj(4, rect);
	  }
	  Tcl_SetObjResult(interp, Tcl_NewListObj(count, retvals));
	  Tcl_Free((char *)retvals);
	}
	if (rects) {
	  XFree(rects);
	}
	return TCL_OK;
      } else if (Tcl_GetIndexFromObj(interp, objv[4], operations,
				     "operation", 0, &op) != TCL_OK) {
	return TCL_ERROR;
      } else {
	XRectangle *rects = (XRectangle *)
	  Tcl_Alloc(sizeof(XRectangle)*(objc-5));
	int i;

	for (i=0 ; i<objc-5 ; i++) {
	  Tcl_Obj **coords;
	  int numcoords, x1, y1, x2, y2;
#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((x)<(y)?(y):(x))

	  if (Tcl_ListObjGetElements(interp, objv[i+5],
				     &numcoords, &coords) != TCL_OK) {
	    Tcl_Free((char *)rects);
	    return TCL_ERROR;
	  } else if (numcoords!=4) {
	    Tcl_Free((char *)rects);
	    Tcl_AppendResult(interp, "rectangles should be defined by four"
			     " numbers (x1, y1, x2, and y2)", NULL);
	    return TCL_ERROR;
	  } else if (Tcl_GetIntFromObj(interp, coords[0], &x1) != TCL_OK ||
		     Tcl_GetIntFromObj(interp, coords[1], &y1) != TCL_OK ||
		     Tcl_GetIntFromObj(interp, coords[2], &x2) != TCL_OK ||
		     Tcl_GetIntFromObj(interp, coords[3], &y2) != TCL_OK) {
	    Tcl_Free((char *)rects);
	    return TCL_ERROR;
	  }
	  rects[i].x = min(x1,x2);
	  rects[i].y = min(y1,y2);
	  rects[i].width  = max(x1-x2, x2-x1) + 1;
	  rects[i].height = max(y1-y2, y2-y1) + 1;
	}

	if (kind==0 || kind==2) {
	  XShapeCombineRectangles(dpy, w, ShapeBounding, /* What's going on */
				  0, 0,			 /* Offset - unused */
				  rects, objc-5,	 /* The rectangles */
				  opmap[op],		 /* Set operation */
				  Unsorted);		 /* Rects unordered */
	  if (isTop && applyOperationToToplevelParent) {
	    Window parent = getXParent(dpy, w);

	    if (parent!=None) {
	      XShapeCombineRectangles(dpy, parent, ShapeBounding, 0, 0,
				      rects, objc-5, opmap[op], Unsorted);
	    }
	  }
	}
	if (kind==1 || kind==2) {
	  XShapeCombineRectangles(dpy, w, ShapeClip, 0, 0, rects, objc-5,
				  opmap[op], Unsorted);
	  if (isTop && applyOperationToToplevelParent) {
	    Window parent = getXParent(dpy, w);

	    if (parent!=None) {
	      XShapeCombineRectangles(dpy, parent, ShapeClip, 0, 0,
				      rects, objc-5, opmap[op], Unsorted);
	    }
	  }
	}
	Tcl_Free((char *)rects);
	return TCL_OK;
      }
    }

  case SHAPE_VERSION:
    /* shape version */
    if (objc!=2) {
      Tcl_WrongNumArgs(interp, 1, objv, "version");
      return TCL_ERROR;
    } else {
      int major=-1, minor=-1;
      char buffer[64];
      Status s = XShapeQueryVersion(dpy, &major, &minor);

      if (s==True) {
	sprintf(buffer, "%d.%d", major, minor);
	Tcl_AppendResult(interp, buffer, NULL);
      }
      return TCL_OK;
    }

  default:
    panic("switch fallthrough");
  }
}

int
Shape_Init(interp)
     Tcl_Interp *interp;
{
  Tk_Window tkwin = Tk_MainWindow(interp);
  int evBase, errBase;

  if (XShapeQueryExtension(Tk_Display(tkwin), &evBase, &errBase) == False) {
    Tcl_AppendResult(interp, "shaped window extension not supported", NULL);
    return TCL_ERROR;
  }

  Tcl_CreateObjCommand(interp, "shape", foo, tkwin, NULL);
  return TCL_OK;
}
